# Copyright 2024 Open Source Robotics Foundation, Inc.
# Licensed under the Apache License, Version 2.0

from pathlib import Path
import sys
import sysconfig

from colcon_core.location import get_relative_package_index_path
from colcon_core.package_descriptor import PackageDescriptor
from colcon_installed_package_information.package_augmentation.colcon_index \
    import ColconIndexPackageAugmentation
from colcon_installed_package_information.package_augmentation.python \
    import InstalledPythonPackageAugmentation
import pytest


@pytest.fixture
def mock_egg_package(tmp_path):
    lib_dir = Path(sysconfig.get_path('purelib', vars={'base': str(tmp_path)}))
    pkg_dir = lib_dir / 'egg_pkg-4.5.6.egg-info'

    pkg_dir.mkdir(parents=True)
    (pkg_dir / 'PKG-INFO').write_text(
        'Metadata-Version: 2.1\n'
        'Name: egg-pkg\n'
        'Version: 4.5.6\n'
    )
    (pkg_dir / 'requires.txt').write_text(
        'wheel-pkg\n'
    )

    desc = PackageDescriptor(str(tmp_path))
    desc.name = 'egg-pkg'
    desc.type = 'installed'
    yield desc


@pytest.fixture
def mock_wheel_package(tmp_path):
    lib_dir = Path(sysconfig.get_path('purelib', vars={'base': str(tmp_path)}))
    pkg_dir = lib_dir / (
        'wheel_pkg-1.2.3.'
        f'py{sys.version_info.major}.{sys.version_info.minor}.dist-info'
    )

    pkg_dir.mkdir(parents=True)
    (pkg_dir / 'METADATA').write_text(
        'Metadata-Version: 2.1\n'
        'Name: wheel-pkg\n'
        'Version: 1.2.3\n'
    )

    desc = PackageDescriptor(str(tmp_path))
    desc.name = 'wheel-pkg'
    desc.type = 'installed'
    yield desc


@pytest.fixture
def mock_colcon_index_package(tmp_path):
    index_dir = tmp_path / get_relative_package_index_path()

    index_dir.mkdir(parents=True)
    (index_dir / 'colcon-index-pkg').write_text(
        'some-other-pkg',
    )

    desc = PackageDescriptor(str(tmp_path))
    desc.name = 'colcon-index-pkg'
    desc.type = 'installed'
    yield desc


def test_colcon_index(mock_colcon_index_package):
    augmentation_extension = ColconIndexPackageAugmentation()

    augmentation_extension.augment_packages([
        mock_colcon_index_package,
    ])

    assert mock_colcon_index_package.name == 'colcon-index-pkg'
    assert mock_colcon_index_package.type == 'installed.colcon'
    assert 'some-other-pkg' in mock_colcon_index_package.dependencies['run']


@pytest.mark.parametrize('augmentation_extension_class', [
    ColconIndexPackageAugmentation,
    InstalledPythonPackageAugmentation,
])
def test_no_installed(tmp_path, augmentation_extension_class):
    augmentation_extension = augmentation_extension_class()

    desc = PackageDescriptor(str(tmp_path))
    desc.name = 'cmake-package'
    desc.type = 'cmake'

    augmentation_extension.augment_packages([desc])

    assert desc.name == 'cmake-package'
    assert desc.type == 'cmake'


@pytest.mark.parametrize('augmentation_extension_class', [
    ColconIndexPackageAugmentation,
    InstalledPythonPackageAugmentation,
])
def test_not_found(tmp_path, augmentation_extension_class):
    augmentation_extension = augmentation_extension_class()

    desc = PackageDescriptor(str(tmp_path))
    desc.name = 'no-such-package'
    desc.type = 'installed'

    augmentation_extension.augment_packages([desc])

    assert desc.name == 'no-such-package'
    assert desc.type == 'installed'


def test_python(mock_egg_package, mock_wheel_package):
    augmentation_extension = InstalledPythonPackageAugmentation()

    augmentation_extension.augment_packages([
        mock_egg_package,
        mock_wheel_package,
    ])

    assert mock_egg_package.name == 'egg-pkg'
    assert mock_egg_package.type == 'installed.python'
    assert mock_egg_package.metadata.get('version') == '4.5.6'
    assert 'wheel-pkg' in mock_egg_package.dependencies['run']

    assert mock_wheel_package.name == 'wheel-pkg'
    assert mock_wheel_package.type == 'installed.python'
    assert mock_wheel_package.metadata.get('version') == '1.2.3'
    assert not any(mock_wheel_package.dependencies.items())



# Copyright 2022 Open Source Robotics Foundation, Inc.
# Licensed under the Apache License, Version 2.0

import os

from colcon_core.dependency_descriptor import DependencyDescriptor
from colcon_core.location import get_relative_package_index_path
from colcon_core.package_augmentation \
    import PackageAugmentationExtensionPoint
from colcon_core.plugin_system import satisfies_version


class ColconIndexPackageAugmentation(PackageAugmentationExtensionPoint):
    """
    Augment installed packages with information from a colcon index.

    Only packages of the `installed` type are considered.
    """

    # This extension adds dependency information for installed packages more
    # efficiently than the Python extension, so it should have a higher
    # higher priority than it.
    PRIORITY = 120

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            PackageAugmentationExtensionPoint.EXTENSION_POINT_VERSION,
            '^1.0')

    def augment_package(  # noqa: D102
        self, desc, *, additional_argument_names=None
    ):
        if desc.type != 'installed':
            return

        marker_file = desc.path / get_relative_package_index_path() / desc.name
        if not marker_file.is_file():
            return

        with marker_file.open() as f:
            raw_deps = f.read().split(os.pathsep)
        desc.type = 'installed.colcon'
        desc.dependencies['run'].update(
            DependencyDescriptor(dep) for dep in raw_deps)



# Copyright 2022 Open Source Robotics Foundation, Inc.
# Licensed under the Apache License, Version 2.0

from pathlib import Path

from colcon_core.package_descriptor import PackageDescriptor
from colcon_core.package_discovery import PackageDiscoveryExtensionPoint
from colcon_core.plugin_system import satisfies_version
from colcon_core.prefix_path import get_chained_prefix_path
from colcon_core.shell import find_installed_packages
from colcon_installed_package_information.package_discovery import logger


class PrefixPathPackageDiscovery(PackageDiscoveryExtensionPoint):
    """Discover packages in chained prefix paths."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            PackageDiscoveryExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def has_parameters(self, *, args):  # noqa: D102
        return False

    def discover(self, *, args, identification_extensions):  # noqa: D102
        descs = set()

        for priority, prefix_path in enumerate(get_chained_prefix_path()):
            packages = find_installed_packages(Path(prefix_path))
            if packages is None:
                continue
            num_packages = len(packages)
            logger.debug('Found {num_packages} installed packages in '
                         '{prefix_path}'.format_map(locals()))
            for pkg, path in (packages or {}).items():
                desc = PackageDescriptor(path)
                desc.name = pkg
                desc.type = 'installed'
                desc.metadata['override_priority'] = priority
                descs.add(desc)

        return descs



# Copyright 2022 Scott K Logan
# Licensed under the Apache License, Version 2.0

import logging

from colcon_core.logging import colcon_logger
from colcon_installed_package_information.package_augmentation \
    import augment_packages
from colcon_installed_package_information.package_discovery \
    import discover_packages
from colcon_installed_package_information.package_identification \
    import get_package_identification_extensions


def main():
    """
    PEP 338 entry point used for testing this module.

    This function will enumerate all of the installed packages and print them
    to the console in a similar way to the `colcon list` command. Because this
    functionality provides only limited utility outside of debugging, it is
    not exposed as a full colcon verb.
    """
    colcon_logger.setLevel(logging.WARNING)

    # Enumerate the underlay packages.
    identification_extensions = get_package_identification_extensions()
    descriptors = discover_packages(None, identification_extensions)
    augment_packages(descriptors)

    colcon_logger.debug('Found {} packages'.format(len(descriptors)))

    lines = []
    for pkg in sorted(descriptors, key=lambda pkg: pkg.name):
        lines.append(pkg.name + '\t' + str(pkg.path) + '\t(%s)' % pkg.type)

    for line in lines:
        print(line)


if __name__ == '__main__':
    main()



# Copyright 2022 Open Source Robotics Foundation, Inc.
# Licensed under the Apache License, Version 2.0

import sysconfig

try:
    from importlib.metadata import Distribution
except ImportError:
    # TODO: Drop this when dropping Python 3.7 support
    from importlib_metadata import Distribution

from colcon_core.package_augmentation \
    import PackageAugmentationExtensionPoint
from colcon_core.package_augmentation.python \
    import create_dependency_descriptor
from colcon_core.plugin_system import satisfies_version


class InstalledPythonPackageAugmentation(PackageAugmentationExtensionPoint):
    """
    Augment installed packages with Python distribution information.

    Only packages of the `installed` type are considered.
    """

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            PackageAugmentationExtensionPoint.EXTENSION_POINT_VERSION,
            '^1.0')

    def augment_packages(  # noqa: D102
        self, descs, *, additional_argument_names=None
    ):
        descs = {d for d in descs if d.type == 'installed'}
        if not descs:
            return

        for desc in descs:
            dists = Distribution.discover(
                path=_enumerate_python_dirs(str(desc.path)),
                name=desc.name)
            try:
                dist = next(iter(dists))
            except StopIteration:
                continue

            if not desc.metadata.get('version'):
                version = dist.version
                if version:
                    desc.metadata['version'] = version
            desc.type = 'installed.python'
            # TODO: We should find a clean way to exclude test-oriented extras
            #       from this enumeration.
            requires = dist.requires
            if requires is not None:
                desc.dependencies['run'].update(
                    create_dependency_descriptor(str(req))
                    for req in requires)


def _enumerate_python_dirs(prefix):
    get_path_vars = {'base': prefix, 'platbase': prefix}
    yield sysconfig.get_path('purelib', vars=get_path_vars)
    yield sysconfig.get_path('platlib', vars=get_path_vars)



# Copyright 2022 Open Source Robotics Foundation, Inc.
# Licensed under the Apache License, Version 2.0

from colcon_core.plugin_system import instantiate_extensions
from colcon_core.plugin_system import order_extensions_grouped_by_priority


def get_package_identification_extensions():
    """
    Get the available package identification extensions.

    The extensions are grouped by their priority and each group is ordered by
    the entry point name.

    :rtype: OrderedDict
    """
    extensions = instantiate_extensions(__name__)
    for name, extension in extensions.items():
        extension.PACKAGE_IDENTIFICATION_NAME = name
    return order_extensions_grouped_by_priority(extensions)



# Copyright 2022 Open Source Robotics Foundation, Inc.
# Licensed under the Apache License, Version 2.0

from colcon_core.package_augmentation \
    import augment_packages as augment_packages_impl
from colcon_core.plugin_system import instantiate_extensions
from colcon_core.plugin_system import order_extensions_by_priority


def get_package_augmentation_extensions():
    """
    Get the available package augmentation extensions.

    The extensions are ordered by their priority and entry point name.

    :rtype: OrderedDict
    """
    extensions = instantiate_extensions(__name__)
    for name, extension in extensions.items():
        extension.PACKAGE_AUGMENTATION_NAME = name
    return order_extensions_by_priority(extensions)


def augment_packages(
    descs, *, additional_argument_names=None, augmentation_extensions=None
):
    """
    Augment package descriptors with additional information.

    :param descs: the packages
    :type descs: set of
      :py:class:`colcon_core.package_descriptor.PackageDescriptor`
    """
    if augmentation_extensions is None:
        augmentation_extensions = get_package_augmentation_extensions()

    augment_packages_impl(
        descs, additional_argument_names=additional_argument_names,
        augmentation_extensions=augmentation_extensions)



# Copyright 2022 Open Source Robotics Foundation, Inc.
# Licensed under the Apache License, Version 2.0

from colcon_core.logging import colcon_logger
from colcon_core.package_discovery import \
    add_package_discovery_arguments as add_package_discovery_arguments_impl
from colcon_core.package_discovery import \
    discover_packages as discover_packages_impl
from colcon_core.plugin_system import instantiate_extensions
from colcon_core.plugin_system import order_extensions_by_priority

logger = colcon_logger.getChild(__name__)


def get_package_discovery_extensions():
    """
    Get the available package discovery extensions.

    The extensions are ordered by their priority and entry point name.

    :rtype: OrderedDict
    """
    extensions = instantiate_extensions(__name__)
    for name, extension in extensions.items():
        extension.PACKAGE_DISCOVERY_NAME = name
    return order_extensions_by_priority(extensions)


def add_package_discovery_arguments(parser, *, extensions=None):
    """
    Add the command line arguments for the package discovery extensions.

    :param parser: The argument parser
    :param extensions: The package discovery extensions to use, if `None` is
      passed use the extensions provided by
      :function:`get_package_discovery_extensions`
    """
    if extensions is None:
        extensions = get_package_discovery_extensions()
    add_package_discovery_arguments_impl(parser, extensions=extensions)


def discover_packages(
    args, identification_extensions, *, discovery_extensions=None
):
    """
    Discover and identify packages.

    All discovery extensions which report to have parameters are being used to
    discover packages.
    If none report to have parameters all discovery extensions are being used
    but only the one with a default value should discover packages.
    Each discovery extension uses the passed identification extensions to check
    each potential location for the existence of a package.

    :param args: The parsed command line arguments
    :param identification_extensions: The package identification extensions to
      pass to each invocation of
      :function:`PackageDiscoveryExtensionPoint.discover`
    :param discovery_extensions: The package discovery extensions to use, if
      `None` is passed use the extensions provided by
      :function:`get_package_discovery_extensions`
    :returns: set of
      :py:class:`colcon_core.package_descriptor.PackageDescriptor`
    :rtype: set
    """
    if discovery_extensions is None:
        discovery_extensions = get_package_discovery_extensions()
    return discover_packages_impl(
        args, identification_extensions,
        discovery_extensions=discovery_extensions)
